from itertools import groupby
from typing import List, Literal, Union
from .api_base_model import ApiBaseModelWithIdNameLabelAndDesc
from .schedule_timeline_exit import ScheduleTimelineExit
from .scheduled_instance import ScheduledActivityInstance, ScheduledDecisionInstance
from .timing import Timing
from .duration import Duration


class ScheduleTimeline(ApiBaseModelWithIdNameLabelAndDesc):
    mainTimeline: bool
    entryCondition: str
    entryId: str
    exits: List[ScheduleTimelineExit] = []
    timings: List[Timing] = []
    instances: List[Union[ScheduledActivityInstance, ScheduledDecisionInstance]] = []
    plannedDuration: Union[Duration, None] = None
    instanceType: Literal["ScheduleTimeline"]

    def first_timepoint(self) -> ScheduledActivityInstance:
        return self.instances[0] if self.instances else None

    def find_timepoint(self, id: str) -> ScheduledActivityInstance:
        return next((x for x in self.instances if x.id == id), None)

    def timepoint_list(self) -> list:
        return self.instances

    def find_timing_from(self, id: str) -> Timing:
        return next(
            (x for x in self.timings if x.relativeFromScheduledInstanceId == id), None
        )

    # @todo Nice to get StudyDesign type hint included but circular ...
    def soa(self, study_design) -> list:
        # Activities
        activity_order = study_design.activity_list()

        # Epochs and Visits
        ai = []
        timepoints = self.timepoint_list()
        for timepoint in timepoints:
            timing = self.find_timing_from(timepoint.id)
            encounter = study_design.find_encounter(timepoint.encounterId)
            epoch = study_design.find_epoch(timepoint.epochId)
            # print(f"TPT: {timepoint}")
            # print(f"ENCOUNTER: {encounter}")
            # print(f"EPOCH: {epoch}")
            entry = {
                "instance": timepoint,
                "timing": timing,
                "epoch": epoch,
                "encounter": encounter,
            }
            ai.append(entry)

        # Blank row
        visit_row = {}
        for item in ai:
            visit_row[item["instance"].id] = ""

        # Activities
        activities = {}
        for activity in activity_order:
            for timepoint in timepoints:
                if activity.id in timepoint.activityIds:
                    if activity.name not in activities:
                        activities[activity.name] = visit_row.copy()
                    activities[activity.name][timepoint.id] = "X"

        # Form results
        labels = []
        for item in ai:
            label = (
                item["encounter"].label if item["encounter"] else item["instance"].label
            )
            labels.append(label)
        results = []
        results.append(
            [""] + [item["epoch"].label if item["epoch"] else "&nbsp;" for item in ai]
        )
        results.append([""] + labels)
        results.append(
            [""]
            + [item["instance"].label if item["instance"] else "&nbsp;" for item in ai]
        )
        results.append(
            [""]
            + [
                item["timing"].windowLabel if item["timing"] else "&nbsp;"
                for item in ai
            ]
        )
        for activity in activity_order:
            if activity.name in activities:
                data = activities[activity.name]
                label = activity.label if activity.label else activity.name
                results.append(
                    [{"name": label, "id": activity.id}] + list(data.values())
                )

        # Format as HTML
        lines = []
        lines.append('<table class="soa-table table">')
        lines.append("<thead>")
        lines.append('<tr class="table-active">')
        epochs = [[i, len([*group])] for i, group in groupby(results[0])]
        text = '<td><p class="m-0 p-0"><small>&nbsp;</small></p></td>'
        for epoch in epochs[1:]:
            text += f'<td class="text-center" colspan="{epoch[1]}"><p class="m-0 p-0"><small>{epoch[0]}</small></p></td>'
        lines.append(text)
        lines.append("</tr>")
        for result in results[1:4]:
            lines.append('<tr class="table-active">')
            lines.append('<td><p class="m-0 p-0"><small>&nbsp;</small></p></td>')
            for cell in result[1:]:
                lines.append(
                    f'<td class="text-center"><p class="m-0 p-0"><small>{cell}</small></p></td>'
                )
            lines.append("</tr>")
        lines.append("</thead>")
        lines.append("<tbody>")
        for result in results[4:]:
            lines.append("<tr>")
            x = result[0]["name"]
            lines.append(
                f'<td class="m-0 p-0"><p class="m-0 p-0"><small>{x}</small></p></td>'
            )
            for cell in result[1:]:
                lines.append(
                    f'<td class="m-0 p-0 text-center"><p class="m-0 p-0"><small>{cell}</small></p></td>'
                )
            lines.append("</tr>")
        lines.append("</tbody>")
        lines.append("</table>")
        return ("\n").join(lines)



from src.usdm4 import USDM4
from tests.helpers.files import write_json_file, read_json_file

SAVE = False


def dump_validation_result(result):
    print(
        f"RESULT: {[v for k, v in result._items.items() if v['status'] not in ['Not Implemented', 'Success']]}"
    )


def test_validate(tmp_path):
    result = USDM4().validate("tests/test_files/test_validate.json")
    dump_validation_result(result)
    assert result.passed_or_not_implemented()


def test_validate_error():
    result = USDM4().validate("tests/test_files/test_validate_error.json")
    dump_validation_result(result)
    assert not result.passed_or_not_implemented()


def test_example_1():
    test_file = "tests/test_files/package/example_1.json"
    result = USDM4().validate(test_file)
    assert not result.passed_or_not_implemented()


def test_example_2():
    test_file = "tests/test_files/package/example_2.json"
    result = USDM4().validate(test_file)
    dump_validation_result(result)
    assert result.passed_or_not_implemented()


def test_minimum():
    result = USDM4().minimum("Test Study", "SPONSOR-1234", "1")
    result.study.id = "FAKE-UUID"
    if SAVE:
        write_json_file(None, "test_minimum_expected.json", result.to_json())
    expected = read_json_file(None, "test_minimum_expected.json")
    assert result.to_json() == expected



import json
import pathlib
from usdm4.rules.rules_validation import RulesValidation4
from usdm3.rules.rules_validation_results import RulesValidationResults
from usdm4.api.wrapper import Wrapper
from usdm4.convert.convert import Convert
from usdm4.builder.builder import Builder


class USDM4:
    def __init__(self):
        self.root = self._root_path()
        self.validator = RulesValidation4(self.root)

    def validate(self, file_path: str) -> RulesValidationResults:
        return self.validator.validate(file_path)

    def convert(self, file_path: str) -> Wrapper:
        with open(file_path, "r") as file:
            data = json.load(file)
        return Convert.convert(data)

    def builder(self) -> Builder:
        return Builder(self.root)

    def minimum(self, study_name: str, sponsor_id: str, version: str) -> Wrapper:
        return Builder(self.root).minimum(study_name, sponsor_id, version)

    def from_json(self, data: dict) -> Wrapper:
        return Wrapper.model_validate(data)

    def _root_path(self) -> str:
        return pathlib.Path(__file__).parent.resolve()



from usdm3.rules.library.rule_template import RuleTemplate


class RuleDDF00113(RuleTemplate):
    """
    DDF00113: An agent administration's frequency must be specified according to the extensible Frequency (C71113) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).

    Applies to: AgentAdministration
    Attributes: frequency
    """

    def __init__(self):
        super().__init__(
            "DDF00113",
            RuleTemplate.ERROR,
            "An agent administration's frequency must be specified according to the extensible Frequency (C71113) SDTM codelist (e.g. an entry with a code or decode used from the codelist should be consistent with the full entry in the codelist).",
        )

    def validate(self, config: dict) -> bool:
        raise NotImplementedError("rule is not implemented")



import json
import enum
import datetime
from uuid import UUID
from typing import Union, List
from pydantic import BaseModel, Field
from .extension import ExtensionAttribute


# Example, see https://stackoverflow.com/questions/10252010/serializing-class-instance-to-json
def _serialize_as_json(obj):
    if isinstance(obj, enum.Enum):
        return obj.value
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, UUID):
        return str(obj)
    else:
        return obj.__dict__


def _serialize_as_json_with_type(obj):
    if isinstance(obj, enum.Enum):
        return obj.value
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, UUID):
        return str(obj)
    else:
        result = obj.__dict__
        result["_type"] = obj.__class__.__name__
        return result


class ApiBaseModel(BaseModel):
    def __init__(self, *args, **kwargs):
        kwargs["instanceType"] = self.__class__.__name__
        super().__init__(*args, **kwargs)

    def to_json(self):
        return json.dumps(self, default=_serialize_as_json)

    def to_json_with_type(self):
        return json.dumps(self, default=_serialize_as_json_with_type)


class ApiBaseModelWithIdOnly(ApiBaseModel):
    id: str = Field(min_length=1)


class ApiBaseModelWithId(ApiBaseModelWithIdOnly):
    extensionAttributes: List[ExtensionAttribute] = []


class ApiBaseModelWithIdAndDesc(ApiBaseModelWithId):
    description: Union[str, None] = None


class ApiBaseModelWithIdAndName(ApiBaseModelWithId):
    name: str = Field(min_length=1)


class ApiBaseModelWithIdNameAndLabel(ApiBaseModelWithIdAndName):
    label: Union[str, None] = None

    def label_name(self) -> str:
        return self.label if self.label else self.name


class ApiBaseModelWithIdNameLabelAndDesc(ApiBaseModelWithIdNameAndLabel):
    description: Union[str, None] = None


class ApiBaseModelWithIdNameAndDesc(ApiBaseModelWithIdAndName):
    description: Union[str, None] = None



class Library:
    def __init__(self, root_path: str):
        self.system = "ISO 639-1"
        self.version = "2007"

    def load(self) -> None:
        pass

    def decode(self, code: str) -> str | None:
        # Just support "en" for the moment
        return "English" if code == "en" else None
